<?php
/**
 * Voicer
 * Voicer converts text into human-like speech in more than 295 voices across 50+ languages and variants.
 * Exclusively on https://1.envato.market/voicer
 *
 * @encoding        UTF-8
 * @version         3.2.3
 * @copyright       (C) 2018 - 2022 Merkulove ( https://merkulov.design/ ). All rights reserved.
 * @license         Envato License https://1.envato.market/KYbje
 * @contributors    Nemirovskiy Vitaliy (nemirovskiyvitaliy@gmail.com), Dmitry Merkulov (dmitry@merkulov.design)
 * @support         help@merkulov.design
 **/

namespace Merkulove\Voicer;

use DOMDocument;
use DOMXPath;

use Merkulove\Voicer\Unity\Settings;
use Merkulove\Voicer\Unity\Plugin;
use Merkulove\Voicer\Unity\TabAssignments;
use Merkulove\Voicer\Unity\UI;

use Google\ApiCore\ApiException;
use Google\Cloud\TextToSpeech\V1\AudioConfig;
use Google\Cloud\TextToSpeech\V1\AudioEncoding;
use Google\Cloud\TextToSpeech\V1\SynthesisInput;
use Google\Cloud\TextToSpeech\V1\TextToSpeechClient;
use Google\Cloud\TextToSpeech\V1\VoiceSelectionParams;

/** Exit if accessed directly. */
if ( ! defined( 'ABSPATH' ) ) {
	header( 'Status: 403 Forbidden' );
	header( 'HTTP/1.1 403 Forbidden' );
	exit;
}

/**
 * SINGLETON: Caster class contain main plugin logic.
 *
 * @since 1.0.0
 *
 **/
final class Caster {

	/**
	 * The one true Caster.
	 *
     * @since 1.0.0
     * @access private
	 * @var Caster
	 **/
	private static $instance;

    /**
     * Setup the plugin.
     *
     * @since 1.0.0
     * @access public
     *
     * @return void
     **/
    public function setup() {

        /** Define hooks that runs on both the front-end as well as the dashboard. */
        $this->both_hooks();

        /** Define public hooks. */
        $this->public_hooks();

        /** Define admin hooks. */
        $this->admin_hooks();

    }

    /**
     * Define hooks that runs on both the front-end as well as the dashboard.
     *
     * @since 1.0.0
     * @access private
     *
     * @return void
     **/
    private function both_hooks() {

        /** Hooks which should work everywhere in backend and frontend. */

        /** Ajax on the Viewer-Facing Side. */
        add_action( 'wp_ajax_nopriv_voicer_gspeak', [ $this, 'gspeak' ] );
        add_action( 'wp_ajax_voicer_gspeak', [ $this, 'gspeak' ] );

        /** Adds all the necessary shortcodes. */
        Shortcodes::get_instance();

    }

    /**
     * @param $pattern
     * @param $array
     *
     * @return int
     */
    private function preg_array_key_exists( $pattern, $array ) {

        // extract the keys.
        $keys = array_keys( $array );

        // convert the preg_grep() returned array to int..and return.
        // the ret value of preg_grep() will be an array of values
        // that match the pattern.
        return (int) preg_grep( $pattern, $keys );

    }

    /**
     * Validate and return language code
     * @param $options
     *
     * @return int|mixed|string|string[]
     */
    private function get_lang_code( $options ) {

        // Get locale and list of languages
        $languages = Config::$languages;

        // Replace _ by - in the language code
        $language_code = str_replace( '_', '-', $options[ 'html_lang' ] );

        // Get basic language from locale
        $base_locale = ! strstr( $language_code, '-' ) ?
            $language_code : // Base language without dash
            strstr( $language_code, '-', true ); // Base language with dash

        /** Check is voice esist in the Google Voices list */
        if ( array_key_exists( $language_code, $languages ) ) {

            return $language_code;

        /** Check is basic languages from locale exist in the Google Vocies list */
        } else if ( $this->preg_array_key_exists( '/(' . $base_locale . '-)/', $languages ) ) {

            // Find firs language with similar base lanhuage code
            foreach ( $languages as $lang_key => $lang_value ) {
                if ( strpos( $lang_key, $base_locale ) === 0 ) {
                    return $lang_key;
                    break;
                }
            }

        /** No exext mach or similar languages */
        } else {

            return $options[ 'language_code' ];

        }

    }

    /**
     * Returs language properties
     * @param $options
     *
     * @return array
     */
    private function get_language( $options ) {

        /** Prepare language and language code */
        if ( ! empty( $options['html_voice'] ) ) {

	        $language_code = $options[ 'html_lang' ];
	        $voice_code = $options[ 'html_voice' ];

        } else if ( 'on' === $options[ 'multi' ] ) {

	        $language_code = $this->get_lang_code( $options );
	        $voice_code = $language_code . '-Standard-A';

        } else {

            $language_code = $options['language-code'];
            $voice_code = $options['language'];

        }

        return [ 'language_code' => $language_code, 'voice_code' => $voice_code ];

    }

    /**
     * Register all of the hooks related to the public-facing functionality.
     *
     * @since 1.0.0
     * @access private
     *
     * @return void
     **/
    private function public_hooks() {

        /** Work only on frontend area. */
        if ( is_admin() ) { return; }

        /** Public hooks and filters. */
        add_action( 'wp_enqueue_scripts', [ $this, 'enqueue_frontend_scripts' ] );
        add_action( 'wp_enqueue_scripts', [ $this, 'enqueue_frontend_styles' ] );

    }

    /**
     * Add scripts on front-end
     */
    public function enqueue_frontend_scripts() {

        $options = Settings::get_instance()->options;

        /** Checks if plugin have API Key. */
        if( ! $options['dnd-api-key'] ){ return; }

        /** Checks if plugin should work on this page. */
        if( ! TabAssignments::get_instance()->display() ) { return; }

        /** Enqueue voicer.js as frontend script */
        wp_enqueue_script( 'mdp-voicer', Plugin::get_url() . 'js/voicer' . Plugin::get_suffix() . '.js', [], Plugin::get_version(), true );

        $frontend_options = array(
	        'ajaxurl'           => admin_url( 'admin-ajax.php' ),
	        'nonce'             => wp_create_nonce( 'voicer-nonce' ),
	        'use'               => $options[ 'use' ] === 'on' ? 'true' : 'false',
	        'bgColor'           => $options[ 'bgcolor' ] ?? '#333',
	        'radius'            => $options[ 'radius' ],
	        'iconColor'         => $options[ 'iconcolor' ] ?? '#fff',
	        'playIcon'          => $this->get_svg( $options[ 'play_icon' ] ),
	        'stopIcon'          => $this->get_svg( $options[ 'stop_icon' ] ),
	        'iconSize'          => $options[ 'icon_size' ],
	        'iconMargin'        => $options[ 'icon_margin' ],
	        'highlightDelay'    => $options[ 'shortcode_highlight_delay' ],
            'highlightPosition' => $options[ 'shortcode_highlight_position' ],
            'arrowSize'         => 5,
        );

	    if ( $options[ 'on_select' ] === 'on' ) {

		    $frontend_options[ 'onSelect' ] = $options[ 'on_select' ] === 'on';

	    }

        if ( $options[ 'on_hover' ] === 'on' ) {

	        $frontend_options[ 'onHover' ] = $options[ 'hover_selectors' ];
	        $frontend_options[ 'hoverDelay' ] = $options[ 'hover_delay' ];

        }

        /** In JavaScript, object properties are accessed as mdpVoicer */
        wp_localize_script( 'mdp-voicer', 'mdpVoicer', $frontend_options );

    }

	/**
     * Get svg code of the svg file
	 * @param $file
	 *
	 * @return mixed|string
	 */
    public function get_svg( $file )
    {

	    $file_content = wp_remote_get( Plugin::get_url() . 'images/mdc-icons/' . $file, [
		    'sslverify'  => Settings::get_instance()->options[ 'check_ssl' ] === 'on'
	    ] );

	    return is_array( $file_content ) && isset( $file_content[ 'body' ] ) ? $file_content[ 'body' ] : '';

    }

    /**
     * Add styles on front-end
     */
    public function enqueue_frontend_styles() {

        $options = Settings::get_instance()->options;

        /** Checks if plugin have API Key. */
        if( ! $options['dnd-api-key'] ){ return; }

        /** Checks if plugin should work on this page. */
        if( ! TabAssignments::get_instance()->display() ) { return; }

        /** Add CSS */
        wp_enqueue_style( 'mdp-voicer', Plugin::get_url() . 'css/voicer' . Plugin::get_suffix() . '.css', [], Plugin::get_version() );

        /** Add custom CSS. */
        if ( $options['custom_css'] !== '' ) {
            wp_add_inline_style( 'mdp-voicer', $options['custom_css'] );
        }

    }

    /**
     * Register all of the hooks related to the admin area functionality.
     *
     * @since 1.0.0
     * @access private
     *
     * @return void
     **/
    private function admin_hooks() {

        /** Work only in admin area. */
        if ( ! is_admin() ) { return; }

        /** Add additional scripts to admin area. */
        AdminScripts::get_instance();

        /** Add dashboard widget */
        DashboardWidget::get_instance();

        /** Show admin warning, if we need API Key. */
        if( ! Settings::get_instance()->options['dnd-api-key'] ){
            add_action( 'admin_footer', [ $this, 'key_notice' ] );
        }

    }

	/**
	 * This method used in register_activation_hook
	 * Everything written here will be done during plugin activation.
	 *
	 * @since 1.0.0
	 * @access public
	 */
	public function activation_hook() {

		/** Activation hook */

	}

    /**
     * Ajax front-end action hook here.
     *
     * @throws ApiException
     * @since 1.0.0
     * @access public
     *
     * @return void
     **/
    public function gspeak() {

        /** Security Check. */
        check_ajax_referer( 'voicer-nonce', 'nonce' );

        /** Includes Google Text To Speech classes */
	    require Plugin::get_path() . 'vendor/' . Config::get_instance()->vendor_version() . '/autoload.php';

        /** Get plugin settings. */
        $options = Settings::get_instance()->options;
        $options[ 'html_lang' ] = filter_input( INPUT_POST, 'lang' );
        $options[ 'html_voice' ] = filter_input( INPUT_POST, 'voice' );

        /** Setting custom exception handler. */
        set_exception_handler( [ ErrorHandler::class, 'exception_handler' ] );

        /** Instantiates a client. */
        $client = new TextToSpeechClient();

        /** Sets text to be synthesised. */
        $html = filter_input( INPUT_POST, 'text' );

        /** Remove muted elements by class "voicer-mute" or attribute voicer-mute="". */
        $html = $this->remove_muted_html( $html );

        /** Replace <span voicer-break=""></span> to <break time="200ms"/>. */
        $html = $this->replace_break_tag( $html );

        /** Clean HTML. */
        $html = XMLHelper::get_instance()->clean_html( $html );

        /** Strip all html tags, except SSML tags.  */
        $html = strip_tags( $html, '<p><break><say-as><sub><emphasis>');

        /** Remove the white spaces from the left and right sides.  */
        $html = trim( $html );

        /** Convert HTML entities to their corresponding characters: &quot; => " */
        $html = html_entity_decode( $html );

        /** Sets text to be synthesised. */
        $synthesisInputText = ( new SynthesisInput() )->setSsml( "<speak>" . $html . "</speak>" );

        /** Prepare language configuration */
        $lang = $this->get_language( $options );

        /** Build the voice request, select the language. */
        $voice = (new VoiceSelectionParams())
            ->setLanguageCode( $lang[ 'language_code' ] )
            ->setName( $lang[ 'voice_code' ] );

        /** Effects profile. */
        $effectsProfileId = $options['audio-profile'];

        /** Select the type of audio file you want returned. */
        $audioConfig = ( new AudioConfig() )
            ->setAudioEncoding( AudioEncoding::MP3 )
            ->setEffectsProfileId( [ $effectsProfileId ] )
            ->setSpeakingRate( $options['speaking-rate'] )
            ->setPitch( $options['pitch'] )
            ->setSampleRateHertz( 24000 );

        /** Perform text-to-speech request on the text input with selected voice. */
        $response = $client->synthesizeSpeech( $synthesisInputText, $voice, $audioConfig );
        $audioContent = $response->getAudioContent();

        /** Restore previous exception handler. */
        restore_exception_handler();

        /** The response's audioContent is binary. */
        echo $audioContent;

	    DashboardWidget::get_instance()->countChars( $html, $lang[ 'voice_code' ] );

        wp_die();

    }

    /**
     * Replace <span voicer-break=""></span> to <break time="200ms"/>.
     *
     * @param $html_content string - HTML code.
     *
     * @return string
     * @since 2.0.0
     * @access public
     */
    public function replace_break_tag( $html_content ) {

        /** Create a DOM object. */
        $html = new simple_html_dom();

        /** Load HTML from a string. */
        $html->load( $html_content );

        /** Foreach element wit attribute voicer-break="". */
        foreach ( $html->find( '[voicer-break]' ) as $el ) {

            /** Do replacements. */
            $break_tag = $el->outertext;
            $break_tag = str_replace( '<span voicer-break=""', '<break', $break_tag );

            $el->outertext = $break_tag;
        }

        /** Return result. */
        return $html->save();

    }

    /**
     * Remove muted elements by class "voicer-mute" or attribute voicer-mute="".
     *
     * @param $post_content - Post/Page content.
     *
     * @return string
     * @since 2.0.0
     * @access public
     *
     */
    public function remove_muted_html( $post_content ) {

        /** Hide DOM parsing errors. */
        libxml_use_internal_errors( true );
        libxml_clear_errors();

        /** Load the possibly malformed HTML into a DOMDocument. */
        $dom          = new DOMDocument();
        $dom->recover = true;
        $dom->loadHTML( '<?xml encoding="UTF-8"><body id="repair">' . $post_content . '</body>' ); // input UTF-8.

        $selector = new DOMXPath( $dom );

        /** Remove all elements with voicer-mute="" attribute. */
        foreach( $selector->query( '//*[@voicer-mute]') as $e ) {
            $e->parentNode->removeChild( $e );
        }

        /** Remove all elements with class="voicer-mute". */
        foreach( $selector->query( '//*[contains(attribute::class, "voicer-mute")]' ) as $e ) {
            $e->parentNode->removeChild( $e );
        }

        /** HTML without muted tags. */
        $body = $dom->documentElement->lastChild;

        return trim( XMLHelper::get_instance()->get_inner_html( $body ) );
    }

    /**
     * Show admin warning, if we need API Key.
     *
     * @since 1.0.0
     * @access public
     **/
    public function key_notice() {

        /** Get current screen. */
        $screen = get_current_screen();
        if ( null === $screen ) { return; }

        /** Voicer Settings Page. */
        if ( in_array( $screen->base ,Plugin::get_menu_bases() )  ) {

            /** Render "Before you start" message. */
            UI::get_instance()->render_snackbar(
                esc_html__( 'This plugin uses the Google Cloud Text-to-Speech API Key File. Set up your Google Cloud Platform project before the start.', 'voicer' ),
                'warning', // Type
                -1, // Timeout
                true, // Is Closable
                [ [ 'caption' => 'Get Key File', 'link' => 'https://docs.merkulov.design/about-key-file-for-the-voicer-wordpress-plugin/' ] ] // Buttons
            );

        } else { ?>

            <div class="settings-error notice notice-warning">
                <p><strong><?php esc_html_e( 'Voicer: Before you begin', 'voicer' ); ?></strong></p>
                <p><?php esc_html_e( 'This plugin uses the Cloud Text-to-Speech API. You need to set up your Google Cloud Platform project and authorization before creating audio from text. Visit', 'voicer' ); ?>
                    <a href="https://docs.merkulov.design/about-key-file-for-the-voicer-wordpress-plugin/"
                       target="_blank"><?php esc_html_e( 'Online Documentation', 'voicer' ); ?></a> <?php esc_html_e( 'for more details.', 'voicer' ); ?>
                </p>
            </div>

            <?php

        }

    }

	/**
	 * Main Caster Instance.
	 * Insures that only one instance of Caster exists in memory at any one time.
	 *
	 * @static
     * @since 1.0.0
     * @access public
     *
	 * @return Caster
	 **/
	public static function get_instance() {

		if ( ! isset( self::$instance ) && ! ( self::$instance instanceof self ) ) {

			self::$instance = new self;

		}

		return self::$instance;

	}

}
